#!/usr/bin/python -tt
'''
   Glump is a small program that allows gluing many files together into one.
   It is useful when you have a number of machines and you want to distribute
   a file to them (such as kickstart, firewall rules, etc), but send
   different bits to different machines. Using glump you can make a file from
   many pieces, and even specify per-host tuning parameters and replace
   in-file variables on the fly. It's name comes from the words "glue" and
   "lump" stuck together.
'''
##
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# $Id: glump.py,v 1.29 2006/08/27 17:39:54 skvidal Exp $
# @Author Konstantin Ryabitsev <icon@linux.duke.edu> ($Author: skvidal $)
# @Version $Date: 2006/08/27 17:39:54 $
#

import libxml2
import os
import ConfigParser
import re
import socket

try: 
    from mod_python import apache
except ImportError: 
    pass

if os.path.isdir('/etc/glump'): 
    CFGDIR = '/etc/glump'
else: 
    CFGDIR = os.getcwd()

class Glump:
    '''
    Main glump class. It's not used directly, but is extended
    from other classes, implementing mod_python or cgi access methods.
    '''
    def __init__(self, cfgfile, app, bundleid=None, hostname=None):
        self.app = app
        self.configs = None
        self.files = None
        self.ctype = 'text/plain'
        self.report = 0
        self.comment = 0
        self.csymbol = '#'
        self.vars = {}
        self.params = {}
        self.pieces = [] # [(piecename, missingok)]
        self.dbg = 0
        self.expandvars = 1
        self.out = None
        self._readConfig(cfgfile)
        ##
        # bundleid wins
        #
        if bundleid is not None: 
            self.bundleid = bundleid
            if hostname is not None: 
                self.hostname = hostname
            else: 
                self.hostname = None
        else: self._setBundleIdByHostname(hostname) 
        self._gluePieces()

    def _gluePieces(self):
        '''
        Glue various pieces of text files together.
        '''
        self._readGlueXml()
        if self.comment: 
            out = self._mkCommentHeader()
        else: 
            out = ''
        if self.vars:            
            self.debug('self.params=%s' % self.params)
            for param in self.params.keys():
                if param in self.vars.keys(): 
                    self.vars[param] = self.params[param]
                else:
                    self.vars[param] = self.params[param]
                    
        # expand out %varname% variables inside vars/params, if possible.
        newpvdata = {}
        for pv in self.vars.keys():
            replace_pv = '%%%s%%' % pv
            replace_val = self.vars[pv]
            for this in self.vars.keys():
                thisvalue = self.vars[this]
                self.vars[this] = thisvalue.replace(replace_pv, replace_val)
        
                    
        for (piece, missingok) in self.pieces:
            filepiece = os.path.join(self.files, piece)
            try: 
                fh = open(filepiece, 'r')
            except IOError, exc:
                if missingok:
                    continue
                else:
                    self.errout('Unable to open glue piece file %s: %s' % 
                             (piece, exc))
            else:
                contents = fh.read()
                fh.close()
                
                if self.expandvars and self.vars:
                    try: 
                        contents = contents % self.vars
                    except KeyError, exc:
                        msg = 'Error expanding variables in %s: %s'
                        self.errout(msg % (piece, exc))
                out = out + contents
        self.out = out
        
    def _readConfig(self, cfgfile):
        '''
        Reads in the configuration file.
        '''
        app = self.app
        cp = ConfigParser.ConfigParser()
        if not os.access(cfgfile, os.R_OK):
            self.errout('No such file: %s' % cfgfile)
        try: 
            cp.read(cfgfile)
        except ConfigParser.ParsingError:
            self.errout('Error parsing %s!' % cfgfile)
        try: 
            self.ctype = cp.get(app, 'content-type')
        except ConfigParser.NoSectionError:
            self.errout('No such app defined: %s' % app)
        except ConfigParser.NoOptionError: 
            pass
        try: 
            self.configs = cp.get(app, 'configs')
        except ConfigParser.NoOptionError:
            self.errout('Missing "configs" option for app %s' % app)
        try: 
            self.files = cp.get(app, 'files')
        except ConfigParser.NoOptionError:
            self.errout('Missing "files" option for app %s' % app)
        try: 
            self.report = cp.getboolean(app, 'syslog')
        except ConfigParser.NoOptionError: 
            pass
        try: 
            self.dbg = cp.getboolean(app, 'debug')
        except ConfigParser.NoOptionError: 
            pass
        try: 
            self.expandvars = cp.getboolean(app, 'expandvars')
        except ConfigParser.NoOptionError: 
            pass
        try:
            self.comment = cp.getboolean(app, 'commenthdr')
            self.csymbol = cp.get(app, 'commentsymbol')
        except ConfigParser.NoOptionError: 
            pass

    def _readGlueXml(self):
        '''
        Read and parse the glue.xml file.
        '''
        gluexml = os.path.join(self.configs, 'glue.xml')
        try: 
            doc = libxml2.parseFile(gluexml)
        except libxml2.parserError, exc:
            self.errout('Error parsing %s: %s' % (gluexml, exc))
        root = doc.getRootElement()
        kid = root.children
        while kid:
            if kid.name == 'globals': 
                self._readVars(kid)
            elif kid.name == 'bundle':
                bid = '__unlabeled__'
                props = kid.properties
                expandvars = self.expandvars
                while props:
                    if props.name == 'id': 
                        bid = props.content.strip()
                    elif props.name == 'expandvars':
                        if props.content != 'yes': 
                            expandvars = 0
                    props = props.next
                if bid == self.bundleid:
                    self.debug('expandvars=%s' % expandvars)
                    self.expandvars = expandvars
                    self._getGlueData(kid)
                    doc.freeDoc()
                    return
            kid = kid.next
        self.errout('Bundleid %s not found in %s' % (self.bundleid, gluexml))

    def _getGlueData(self, bnode):
        '''
        Read in the glue data and add members to self.pieces.
        '''
        kid = bnode.children
        while kid:
            if kid.name == 'if' or kid.name == 'ifnot': 
                self._doConditional(kid, self._getGlueData)
            elif kid.name == 'member': 
                self._doMember(kid)
            elif kid.name == 'vars': 
                self._readVars(kid)
            kid = kid.next

    def _doMember(self, membernode):
        '''
           Reads in a <member> node
           '''
        kid = membernode
        piece = kid.content.strip()
        piece = piece.replace('%hostname%', self.hostname)
        piece = piece.replace('%bundleid%', self.bundleid)
        # merge var and param data - param data is greater specificity - 
        # use it if conflict b/t the keys of the two
        pvdata = {}
        for var in self.vars.keys():
            pvdata[var] = self.vars[var]
        for param in self.params.keys():
            pvdata[param] = self.params[param]
        
        
        # replace items in the members with pvdata, if possible
        for pv in pvdata.keys():
            replace_pv = '%%%s%%' % pv
            piece = piece.replace(replace_pv, pvdata[pv])
            
        props = kid.properties
        missingok = 0
        while props:
            if props.name == 'missingok': 
                val = props.content.strip()
                if val == 'yes': 
                    missingok = 1
            props = props.next

        self.pieces.append((piece,missingok))
            
        
    def _doConditional(self, kid, parsingMethod):
        '''
        Handle the conditional statement, such as <if> or <ifnot>
        '''
        props = kid.properties
        conditional = kid.name
        param = None
        value = None
        while props:
            if props.name == 'param': 
                param = props.content
            elif props.name == 'value': 
                value = props.content
            props = props.next
        if param is None or value is None: 
            return
        self.debug('Conditional, param=%s; value=%s' % (param, value))
        if conditional == 'if':
            if param not in self.params.keys():
                ## Condition doesn't satisfy.
                return
            if self.params[param] == value:
                ## Condition satisfies
                parsingMethod(kid)
        else:
            ## Reverse logic to above.
            if param in self.params.keys():
                if self.params[param] == value:
                    ## Condition does NOT satisfy.
                    return
            parsingMethod(kid)

    def _readVars(self, varnode):
        '''
        Read in the <vars> section.
        '''
        kid = varnode.children
        while kid:
            if kid.name == 'var':
                props = kid.properties
                name = None
                while props:
                    if props.name == 'name': 
                        name = props.content.strip()
                    props = props.next
                if name is not None:
                    self.vars[name] = kid.content.strip()
            elif kid.name == 'if' or kid.name == 'ifnot':
                self._doConditional(kid, self._readVars)
            kid = kid.next
        
    def _setBundleIdByHostname(self, hostname):
        '''
        Figures out the bundleid we should be working with by the hostname
        specified in the request.
        '''
        self.hostname = hostname
        hostmatchcfg = os.path.join(self.configs, 'hostmatch.cfg')
        hostmatchlist = self._readHostMatchCfg(hostmatchcfg)
        default = '__default__'
        default_tuning = None
        bundleid = None
        for entry in hostmatchlist:
            tunepairs = None
            try:
                members = entry.split(';')
                # Looks like tuning info
                matchpair = members[0]
                tunepairs = members[1:]
            except ValueError: 
                # No tuning info
                matchpair = entry
            try: 
                (hostregex, bundleid) = matchpair.split('=')
            except ValueError: 
                continue
            bundleid = bundleid.strip()
            hostregex = hostregex.strip()
            if hostregex == '__default__': 
                default = bundleid
                default_tuning = tunepairs
            elif re.compile('^' + hostregex).match(hostname): break
            bundleid = None
        ##
        # See if bundleid includes tuning params.
        #
        if bundleid is None:
            bundleid = default
            tunepairs = default_tuning
        if tunepairs is not None:
            self._readTuningParams(tunepairs)
        self.bundleid = bundleid

    def _readTuningParams(self, tunepairs):
        '''
        Read the tuning parameters from hostmatch.cfg
        '''
        for pair in tunepairs:
            pair = pair.strip()
            if not pair: 
                continue
            try: 
                (name, value) = pair.split('=')
            except ValueError: 
                continue
            self.params[name.strip()] = value.strip()

    def _readHostMatchCfg(self, hostmatchcfg):
        '''
        Read in and parse hostmatch.cfg.
        '''
        try: 
            fh = open(hostmatchcfg, 'r')
        except IOError, exc:
            self.errout('Unable to open %s: %s' % (hostmatchcfg, exc))
        contents = fh.read()
        fh.close()
        ##
        # Remove all comments, strip spaces and remove resulting blank lines,
        # and strip the trailing newline, if any.
        #
        contents = re.sub(re.compile('^#.*\n', re.M), '', contents)
        contents = re.sub('^[ \t]*', '', contents)
        contents = re.sub(re.compile('^\n', re.M), '', contents)
        contents = re.sub('\n$', '', contents)
        ##
        # Split lines into a dictionary
        #
        entries = contents.split('\n')
        return entries

    def _mkCommentHeader(self):
        '''
        Makes a comment header to append to the resulting page, which
        describes parameters used in generating this file. This can
        probably be done better using a different method, so that
        it's not reliant on a comment symbol (e.g. for xml files that's 
        useless).
        '''
        out = '%s Requesting host: %s\n' % (self.csymbol, self.hostname)
        out = out + '%s Matching Bundleid: %s\n' % (self.csymbol, self.bundleid)
        out = out + '%s Glued from: %s\n' % (self.csymbol, self.pieces)
        if self.params:
            out = out + '%s Tuning data: %s\n' % (self.csymbol, self.params)
        #out = out + '%s Generated: %s\n' % (self.csymbol, now)
        out = out + '%s\n' % self.csymbol
        return out

    def _mkReport(self):
        '''
        Reports things into the syslog.
        '''
        if not self.report: 
            return
        sysmsg = ('glued a file for %s using bundleid %s' %
                  (self.hostname, self.bundleid))
        import syslog
        syslog.openlog(self.app)
        syslog.syslog(syslog.LOG_NOTICE, sysmsg)
        syslog.closelog()

    def errout(self, msg):
        'This method is a stub, implemented in extending classes'
        pass

    def publish(self):
        'This method is a stub, implemented in extending classes'
        pass

    def debug(self, msg):
        'This method is a stub, implemented in extending classes'
        pass

##
# Mod_python extension to Glump
#

class ModPythonGlump(Glump):
    '''
    Extends the main Glump class and is to be used with mod_python.
    '''
    def __init__(self, cfgfile, req):
        self.req = req
        try: 
            app = req.form['app']
        except KeyError: 
            app = '__default__'
        bundleid = None
        hostname = None
        try: 
            bundleid = req.form['bundleid']
        except KeyError:
            try: 
                hostname = req.form['hostname']
            except KeyError:
                hostip = req.get_remote_host(apache.REMOTE_NOLOOKUP)
                try: 
                    hostname = socket.gethostbyaddr(hostip)[0]
                except (socket.herror, IndexError): 
                    hostname = hostip
        Glump.__init__(self, cfgfile, app, bundleid=bundleid, hostname=hostname)

    def publish(self):
        'Implement the stub method for publishing'
        self.req.content_type = self.ctype
        return self.out

    def errout(self, msg):
        'Implement the stub method for error reporting'
        apache.log_error('Glump error %s' % msg, apache.APLOG_ERR)
        raise apache.SERVER_RETURN, apache.HTTP_INTERNAL_SERVER_ERROR

    def debug(self, msg):
        'Implement the stub method for printing out debug information'
        if self.dbg:
            apache.log_error('Glump debug:' % msg, apache.APLOG_DEBUG)

##
# CGI-based extension to Glump
#

class CGIGlump(Glump):
    '''
    Extends the main Glump class and is to be used with a CGI interface.
    '''
    def __init__(self, cfgfile, app, bundleid=None, hostname=None):
        self.headerSent = 0
        Glump.__init__(self, cfgfile, app, bundleid=bundleid, hostname=hostname)

    def publish(self):
        'Implement the stub method for publishing'
        self._printHeader()
        sys.stdout.write(self.out)

    def errout(self, msg):
        'Implement the stub method for error reporting'
        self._printHeader()
        sys.stdout.write("%s ERROR: %s\n" % (self.csymbol, msg))
        sys.stdout.write("%s Terminating!\n" % self.csymbol)
        sys.exit(1)
        
    def _printHeader(self):
        '''
        Prints out a CGI header.
        '''
        if self.headerSent: 
            return
        sys.stdout.write("Content-type: %s\n" % self.ctype)
        sys.stdout.write("\n")
        self.headerSent = 1

    def debug(self, msg):
        'Implement the stub method for printing out debug information'
        if self.dbg: 
            print 'debug: %s' % msg

#------------------------------------------------------------------------------
# Web requests
#------------------------------------------------------------------------------

def index(req):
    'Handle mod_python index request'
    cfgfile = os.path.join(CFGDIR, 'glump.conf')
    glump = ModPythonGlump(cfgfile, req)
    return glump.publish()

#------------------------------------------------------------------------------
# CGI Requests
#------------------------------------------------------------------------------
def __cgimain(args):
    'Main code for when we are accessed as a CGI'
    cfgfile = os.path.join(CFGDIR, 'glump.conf')
    hostname = None
    bundleid = None
    app = None
    try:
        gopts, cmds = getopt.getopt(args[1:], 'c:h', #IGNORE:W0612
                                    ['help', 'host=', 'bundleid=', 'app='])
        for o, a in gopts:
            if o == '-c': 
                cfgfile = a
            elif o == '-h' or o == '--help': 
                usage()
            elif o == '--host': 
                hostname = a
            elif o == '--bundleid': 
                bundleid = a
            elif o == '--app': 
                app = a
            else: 
                usage()
    except getopt.error, exc:
        print 'Error parsing commandline args: %s' % exc
        usage()
    if app is None: 
        app = os.path.basename(args[0])
        if app == 'glump.py': 
            app = '__default__'
    if bundleid is None and hostname is None:
        try: 
            hostip = os.environ['REMOTE_ADDR']
        except KeyError: 
            hostip = '127.0.0.1'
        try: 
            hostname = socket.gethostbyaddr(hostip)[0]
        except (socket.herror, IndexError): 
            hostname = hostip
    glump = CGIGlump(cfgfile, app, bundleid=bundleid, hostname=hostname)
    glump.publish()

def usage():
    'Print usage notice and exit'
    print """%s [-c configfile] [--app name] [--host hostname] [--bundleid id]
    Glue a file from a bunch of pieces and return it. Useful for CGIs.
    -c configfile
        Use a configfile instead of 'glue.cfg'
    --app appname
        Act as if you are a certain app. If omitted, it will be the filename
        of this script.
    --host hostname
        Use this hostname instead of trying to resolve $REMOTE_ADDR
    --bundleid bundleid
        Force this bundleid. Wins over hostname if both are specified.
    """ % sys.argv[0]
    sys.exit(0)

if __name__ == '__main__':
    import sys, getopt
    __cgimain(sys.argv)

